# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pyre-unsafe

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals


from nebula2.fbthrift.transport.TTransport import TTransportBase, TTransportException, \
        TServerTransportBase
import os
import errno
import select
import socket
import sys
import time
import warnings

try:
    import fcntl
except ImportError:
    # Windows doesn't have this module
    fcntl = None


def py2_compatible_process_time():
    if sys.version_info.major >= 3 and sys.version_info.minor >= 3:
        return time.process_time()
    else:
        return time.clock()


class ConnectionEpoll:
    """ epoll is preferred over select due to its efficiency and ability to
        handle more than 1024 simultaneous connections """

    def __init__(self):
        self.epoll = select.epoll()
        # TODO should we set any other masks?
        # http://docs.python.org/library/select.html#epoll-objects
        self.READ_MASK = select.EPOLLIN | select.EPOLLPRI
        self.WRITE_MASK = select.EPOLLOUT
        self.ERR_MASK = select.EPOLLERR | select.EPOLLHUP

    def read(self, fileno):
        self.unregister(fileno)
        self.epoll.register(
            fileno, self.READ_MASK | self.ERR_MASK
        )

    def write(self, fileno):
        self.unregister(fileno)
        self.epoll.register(fileno, self.WRITE_MASK)

    def unregister(self, fileno):
        try:
            self.epoll.unregister(fileno)
        except Exception:
            pass

    def process(self, timeout):
        # poll() invokes a "long" syscall that will be interrupted by any signal
        # that comes in, causing an EINTR error.  If this happens, avoid dying
        # horribly by trying again with the appropriately shortened timout.
        process_time = py2_compatible_process_time()

        deadline = process_time + float(timeout or 0)
        poll_timeout = float(timeout or -1)
        while True:
            if timeout is not None and timeout > 0:
                poll_timeout = max(0, deadline - py2_compatible_process_time())
            try:
                msgs = self.epoll.poll(timeout=poll_timeout)
                break
            except IOError as e:
                if e.errno == errno.EINTR:
                    continue
                else:
                    raise

        rset = []
        wset = []
        xset = []
        for fd, mask in msgs:
            if mask & self.READ_MASK:
                rset.append(fd)
            if mask & self.WRITE_MASK:
                wset.append(fd)
            if mask & self.ERR_MASK:
                xset.append(fd)

        return rset, wset, xset


class ConnectionSelect:
    def __init__(self):
        self.readable = set()
        self.writable = set()

    def read(self, fileno):
        if fileno in self.writable:
            self.writable.remove(fileno)
        self.readable.add(fileno)

    def write(self, fileno):
        if fileno in self.readable:
            self.readable.remove(fileno)
        self.writable.add(fileno)

    def unregister(self, fileno):
        if fileno in self.readable:
            self.readable.remove(fileno)
        elif fileno in self.writable:
            self.writable.remove(fileno)

    def registered(self, fileno):
        return fileno in self.readable or fileno in self.writable

    def process(self, timeout):
        # select() invokes a "long" syscall that will be interrupted by any
        # signal that comes in, causing an EINTR error.  If this happens,
        # avoid dying horribly by trying again with the appropriately
        # shortened timout.
        deadline = py2_compatible_process_time() + float(timeout or 0)
        poll_timeout = timeout if timeout is None or timeout > 0 else None
        while True:
            if timeout is not None and timeout > 0:
                poll_timeout = max(0, deadline - py2_compatible_process_time())
            try:
                return select.select(list(self.readable), list(self.writable),
                        list(self.readable), poll_timeout)
            except IOError as e:
                if e.errno == errno.EINTR:
                    continue
                else:
                    raise


class TSocketBase(TTransportBase):
    """Base class for both connected and listening sockets"""
    def __init__(self):
        self.handles = {}

    def _resolveAddr(self, family=None):
        if family is None:
            family = socket.AF_UNSPEC
        if self._unix_socket is not None:
            return [(socket.AF_UNIX, socket.SOCK_STREAM, None, None,
                     self._unix_socket)]
        else:
            ai_flags = 0
            if self.host is None:
                ai_flags |= socket.AI_PASSIVE
            return socket.getaddrinfo(self.host, self.port, family,
                                      socket.SOCK_STREAM, 0,
                                      ai_flags)

    def close(self):
        klist = self.handles.keys() if sys.version_info[0] < 3 else \
            list(self.handles.keys())
        for key in klist:
            self.handles[key].close()
            del self.handles[key]

    def getSocketName(self):
        if not self.handles:
            raise TTransportException(TTransportException.NOT_OPEN,
                'Transport not open')
        return next(iter(self.handles.values())).getsockname()

    def fileno(self):
        if not self.handles:
            raise TTransportException(TTransportException.NOT_OPEN,
                'Transport not open')
        if sys.version_info[0] >= 3:
            return list(self.handles.values())[0].fileno()
        else:
            return self.handles.values()[0].fileno()

    def setCloseOnExec(self, closeOnExec):
        self.close_on_exec = closeOnExec
        for handle in self.handles.values():
            self._setHandleCloseOnExec(handle)

    def _setHandleCloseOnExec(self, handle):
        # Windows doesn't have this module, don't set the handle in this case.
        if fcntl is None:
            return

        flags = fcntl.fcntl(handle, fcntl.F_GETFD, 0)
        if flags < 0:
            raise IOError('Error in retrieving file options')
        if self.close_on_exec:
            fcntl.fcntl(handle, fcntl.F_SETFD, flags | fcntl.FD_CLOEXEC)
        else:
            fcntl.fcntl(handle, fcntl.F_SETFD, flags & ~fcntl.FD_CLOEXEC)


class TSocket(TSocketBase):
    """Connection Socket implementation of TTransport base."""

    def __init__(self, host='localhost', port=9090, unix_socket=None, family=None):
        """Initialize a TSocket

        @param host(str)  The host to connect to.
        @param port(int)  The (TCP) port to connect to.
        @param unix_socket(str)  The filename of a unix socket to connect to.
                                 (host and port will be ignored.)
        @param family(int)  Address family for connection. Ignored if
                            unix_socket is specified.
        """
        TSocketBase.__init__(self)
        self.host = host
        self.port = port
        self.handle = None
        self.family = family
        self._unix_socket = unix_socket
        self._timeout = None
        self.close_on_exec = True
        if not unix_socket:
            self.port = int(self.port)

    def __enter__(self):
        if not self.isOpen():
            self.open()
        return self

    def __exit__(self, type, value, traceback):
        if self.isOpen():
            self.close()

    def setHandle(self, h):
        self.handle = h
        self.handles[h.fileno()] = h

    def getHandle(self):
        return self.handle

    def close(self):
        TSocketBase.close(self)
        self.handle = None

    def isOpen(self):
        return self.handle is not None

    def setTimeout(self, ms):
        if ms is None:
            self._timeout = None
        else:
            self._timeout = ms / 1000.0

        if self.handle is not None:
            self.handle.settimeout(self._timeout)

    def getPeerName(self):
        if not self.handle:
            raise TTransportException(TTransportException.NOT_OPEN,
                'Transport not open')
        return self.handle.getpeername()

    def open(self):
        address = None
        try:
            res0 = self._resolveAddr(self.family)
            for res in res0:
                address = res[4]
                handle = socket.socket(res[0], res[1])
                self.setHandle(handle)
                handle.settimeout(self._timeout)
                self.setCloseOnExec(self.close_on_exec)
                try:
                    handle.connect(address)
                except socket.error:
                    self.close()
                    if res is not res0[-1]:
                        continue
                    else:
                        raise
                break
        except socket.error as e:
            if self._unix_socket:
                msg = 'socket error connecting to path %s: %s' % (
                    self._unix_socket, repr(e))
            else:
                msg = 'socket error connecting to host %s, port %s (%s): %s' % (
                    self.host, self.port, repr(address), repr(e))
            raise TTransportException(TTransportException.NOT_OPEN, msg)

    def read(self, sz):
        try:
            buff = self.handle.recv(sz)
            if len(buff) == 0:
                raise TTransportException(type=TTransportException.END_OF_FILE,
                                          message='TSocket read 0 bytes')
        except socket.error as e:
            raise TTransportException(
                type=TTransportException.END_OF_FILE,
                message='Socket read failed: {}'.format(str(e))
            )
        return buff

    def write(self, buff):
        if not self.handle:
            raise TTransportException(TTransportException.NOT_OPEN,
                    'Transport not open')
        sent = 0
        have = len(buff)
        while sent < have:
            try:
                plus = self.handle.send(buff)
            except socket.error as e:
                raise TTransportException(
                    type=TTransportException.END_OF_FILE,
                    message='Socket write failed: {}'.format(str(e))
                )
            assert plus > 0
            sent += plus
            buff = buff[plus:]

    def flush(self):
        pass

class TServerSocket(TSocketBase, TServerTransportBase):
    """Socket implementation of TServerTransport base."""

    def __init__(self, port=9090, unix_socket=None, family=None, backlog=128):
        """Initialize a TServerSocket

        @param family(int): address family for connections. Ignored if
                            unix_socket is specified.
        @param host(str)  The host to connect to.
        @param port(int)  The (TCP) port to connect to.
        @param unix_socket(str)  The filename of a unix socket to connect to.
                                 (host, port, and family will be ignored.)
        @param backlog(int): maximum number of connections in listen queue.
        """
        TSocketBase.__init__(self)
        self.host = None
        self.port = port
        self._unix_socket = unix_socket
        self.family = family
        self.tcp_backlog = backlog
        self.close_on_exec = True
        if not unix_socket:
            self.port = int(self.port)

        # Since we now rely on select() by default to do accepts across
        # multiple socket fds, we can receive two connections concurrently.
        # In order to maintain compatibility with the existing .accept() API,
        # we need to keep track of the accept backlog.
        self._queue = []

    def __enter__(self):
        if not self.isListening():
            self.listen()
        return self

    def __exit__(self, type, value, traceback):
        if self.isListening():
            self.close()

    def getSocketName(self):
        warnings.warn('getSocketName() is deprecated for TServerSocket.  '
                      'Please use getSocketNames() instead.')
        return TSocketBase.getSocketName(self)

    def getSocketNames(self):
        return [handle.getsockname() for handle in self.handles.values()]

    def fileno(self):
        warnings.warn('fileno() is deprecated for TServerSocket.  '
                      'Please use filenos() instead.')
        return TSocketBase.fileno(self)

    def filenos(self):
        return [handle.fileno() for handle in self.handles.values()]

    def _cleanup_unix_socket(self, addrinfo):
        tmp = socket.socket(addrinfo[0], addrinfo[1])
        try:
            tmp.connect(addrinfo[4])
        except socket.error as err:
            eno, message = err.args
            if eno == errno.ECONNREFUSED:
                os.unlink(addrinfo[4])

    def isListening(self):
        return bool(self.handles)

    def listen(self):
        res0 = self._resolveAddr(self.family)

        for res in res0:
            if res[0] == socket.AF_INET6 and res[4][0] == socket.AF_INET6:
                # This happens if your version of python was built without IPv6
                # support.  getaddrinfo() will return IPv6 addresses, but the
                # contents of the address field are bogus.
                # (For example, see http://bugs.python.org/issue8858)
                #
                # Ignore IPv6 addresses if python doesn't have IPv6 support.
                continue

            # We need remove the old unix socket if the file exists and
            # nobody is listening on it.
            if self._unix_socket:
                self._cleanup_unix_socket(res)

            # Don't complain if we can't create a socket
            # since this is handled below.
            try:
                handle = socket.socket(res[0], res[1])
            except Exception:
                continue
            handle.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            self._setHandleCloseOnExec(handle)

            # Always set IPV6_V6ONLY for IPv6 sockets when not on Windows
            if res[0] == socket.AF_INET6 and sys.platform != 'win32':
                handle.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_V6ONLY, True)

            handle.settimeout(None)
            handle.bind(res[4])
            handle.listen(self.tcp_backlog)

            self.handles[handle.fileno()] = handle

        if not self.handles:
            raise TTransportException("No valid interfaces to listen on!")

    def _sock_accept(self):
        if self._queue:
            return self._queue.pop()

        if hasattr(select, "epoll"):
            poller = ConnectionEpoll()
        else:
            poller = ConnectionSelect()

        for filenos in self.handles.keys():
            poller.read(filenos)

        r, _, x = poller.process(0)

        for fd in r:
            self._queue.append(self.handles[fd].accept())

        if not self._queue:
            raise TTransportException("Accept interrupt without client?")

        return self._queue.pop()

    def accept(self):
        return self._makeTSocketFromAccepted(self._sock_accept())

    def _makeTSocketFromAccepted(self, accepted):
        client, addr = accepted
        result = TSocket()
        result.setHandle(client)
        return result



# --coding:utf-8--
#
# Copyright (c) 2021 vesoft inc. All rights reserved.
#
# This source code is licensed under Apache 2.0 License,
# attached with Common Clause Condition 1.0, found in the LICENSES directory.

class AuthResult(object):
    def __init__(self, session_id, timezone_offset, timezone_name):
        self._session_id = session_id
        self._timezone_offset = timezone_offset
        self._timezone_name = timezone_name

    def get_session_id(self):
        return self._session_id

    def get_timezone_offset(self):
        return self._timezone_offset

    def get_timezone_name(self):
        return self._timezone_name



# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pyre-unsafe

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from base64 import b64encode, b64decode
import json
import sys

from nebula2.fbthrift.protocol.TProtocol import TProtocolBase, TProtocolException
from nebula2.fbthrift.Thrift import TType

JSON_OBJECT_START = b'{'
JSON_OBJECT_END = b'}'
JSON_ARRAY_START = b'['
JSON_ARRAY_END = b']'
JSON_NEW_LINE = b'\n'
JSON_PAIR_SEPARATOR = b':'
JSON_ELEM_SEPARATOR = b','
JSON_BACKSLASH = b'\\'
JSON_BACKSLASH_VALUE = ord(JSON_BACKSLASH)
JSON_STRING_DELIMITER = b'"'
JSON_ZERO_CHAR = b'0'
JSON_TAB = b"  "
JSON_CARRIAGE_RETURN = b'\r'
JSON_SPACE = b' '
TAB = b'\t'

JSON_ESCAPE_CHAR = b'u'
JSON_ESCAPE_PREFIX = b"\\u00"

THRIFT_VERSION_1 = 1

THRIFT_NAN = b"NaN"
THRIFT_INFINITY = b"Infinity"
THRIFT_NEGATIVE_INFINITY = b"-Infinity"

JSON_CHAR_TABLE = [ \
#   0   1    2   3   4   5   6   7    8    9    A   B    C    D   E   F
    0,  0,   0,  0,  0,  0,  0,  0,b'b',b't',b'n',  0,b'f',b'r',  0,  0, \
    0,  0,   0,  0,  0,  0,  0,  0,   0,   0,   0,  0,   0,   0,  0,  0, \
    1,  1,b'"',  1,  1,  1,  1,  1,   1,   1,   1,  1,   1,   1,  1,  1, \
]


JSON_CHARS_TO_ESCAPE = set()
for ch_value, mode in enumerate(JSON_CHAR_TABLE):
    if mode == 1:
        continue
    if sys.version_info[0] == 3:
        JSON_CHARS_TO_ESCAPE.add(chr(ch_value).encode('ascii'))
        JSON_CHARS_TO_ESCAPE.add(chr(ch_value))
    else:
        JSON_CHARS_TO_ESCAPE.add(chr(ch_value))
        JSON_CHARS_TO_ESCAPE.add(chr(ch_value).encode('utf-8'))
JSON_CHARS_TO_ESCAPE.add(JSON_BACKSLASH)
JSON_CHARS_TO_ESCAPE.add(JSON_BACKSLASH.decode('utf-8'))

ESCAPE_CHARS = b"\"\\bfnrt"
ESCAPE_CHAR_VALS = [b'"', b'\\', b'\b', b'\f', b'\n', b'\r', b'\t']

NUMERIC_CHAR = b'+-.0123456789Ee'

WHITESPACE_CHARS = {
    JSON_NEW_LINE,
    TAB,
    JSON_CARRIAGE_RETURN,
    JSON_SPACE,
}

def hexChar(x):
    x &= 0x0f
    return hex(x)[2:]


def hexVal(ch):
    if ch >= '0' and ch <= '9':
        return int(ch) - int('0')
    elif ch >= 'a' and ch <= 'f':
        return int(ch) - int('a') + 10
    raise TProtocolException(TProtocolException.INVALID_DATA,
                             "Unexpected hex value")


class TJSONContext:
    def __init__(self, protocol, indentLevel=0):
        self.indentLevel = indentLevel
        self.protocol = protocol

    def write(self, trans):
        return

    def read(self, reader):
        return

    def escapeNum(self):
        return False

    def writeNewLine(self, trans):
        trans.write(JSON_NEW_LINE)
        self.indent(trans)

    def indent(self, trans):
        trans.write(JSON_TAB * self.indentLevel)


class TJSONPairContext(TJSONContext):
    def __init__(self, protocol, indentLevel=0, isMapPair=False):
        TJSONContext.__init__(self, protocol, indentLevel)
        self.first = True
        self.colon = True
        self.isMapPair = isMapPair
        self.skipColon = False

    def write(self, trans):
        if self.first:
            self.first = False
            self.colon = True
        else:
            if self.colon:
                trans.write(JSON_PAIR_SEPARATOR + b" ")
            else:
                trans.write(JSON_ELEM_SEPARATOR)
                if self.isMapPair:
                    self.writeNewLine(trans)
            self.colon = not self.colon

    def read(self, reader):
        if self.first:
            self.first = False
            self.colon = True
        else:
            self.protocol.skipWhitespace()
            if self.colon:
                if self.skipColon:
                    self.skipColon = False
                else:
                    self.protocol.readJSONSyntaxChar(JSON_PAIR_SEPARATOR)
            else:
                self.protocol.readJSONSyntaxChar(JSON_ELEM_SEPARATOR)
            self.colon = not self.colon

    def escapeNum(self):
        return self.colon

class TJSONListContext(TJSONContext):
    def __init__(self, protocol, indentLevel=0):
        TJSONContext.__init__(self, protocol, indentLevel)
        self.first = True

    def read(self, reader):
        if self.first:
            self.first = False
        else:
            self.protocol.skipWhitespace()
            self.protocol.readJSONSyntaxChar(JSON_ELEM_SEPARATOR)

    def write(self, trans):
        if self.first:
            self.first = False
        else:
            trans.write(JSON_ELEM_SEPARATOR)
            self.writeNewLine(trans)

class LookaheadReader():
    def __init__(self, protocol):
        self.protocol = protocol
        self.hasData = False
        self.data = b''

    def read(self):
        if self.hasData is True:
            self.hasData = False
        else:
            self.data = self.protocol.trans.read(1)
        return self.data

    def peek(self):
        if self.hasData is False:
            self.data = self.protocol.trans.read(1)
            self.hasData = True
        return self.data


class ThriftSpec():
    def __init__(self, spec):
        self.spec = spec
        self.nextSpec = None


class StructSpec(ThriftSpec):
    '''
    Wraps thrift_spec of a thrift struct.
    '''
    def readFieldBegin(self, fname, guess_func):
        field_spec = None
        self.nextSpec = None
        if sys.version_info[0] >= 3:
            fname = fname.decode()
        for s in self.spec:
            if s is not None and s[2] == fname:
                field_spec = s
                break

        if field_spec is not None:
            if field_spec[1] == TType.STRUCT:
                self.nextSpec = StructSpec(field_spec[3][1])
            elif field_spec[1] in (TType.SET, TType.LIST):
                self.nextSpec = ListOrSetSpec(field_spec[3])
            elif field_spec[1] == TType.MAP:
                self.nextSpec = MapSpec(field_spec[3])
            return (fname, field_spec[1], field_spec[0])
        else:
            return (fname, guess_func(), 0)

    def getNextSpec(self):
        return self.nextSpec


class ListOrSetSpec(ThriftSpec):
    '''Wraps a list or set's 2-tuple nested type spec.

    getNextSpec is called in readListBegin to *prepare* the spec of
    the list element which may/may not be used depending on whether
    the list is empty.

    For example, to read list<SomeStruct> the following methods will
    be called:
        readListBegin()
            readStructBegin()
            readStructEnd()
            ...
        readListEnd()
    After readListBegin is called the current spec is still
    ListOrSetSpec and its nextSpec is prepared for its element.
    readStructBegin/End will push/pop the element's StructSpec
    whenever a SomeStruct is read.

    -1 tells the generated code that the size of this list is
    undetermined so it needs to use peekList to detect the end of
    the list.
    '''
    def readListBegin(self):
        self.getNextSpec()
        return (self.spec[0], -1)
    readSetBegin = readListBegin

    def getNextSpec(self):
        if self.nextSpec is None:
            if self.spec[0] == TType.STRUCT:
                self.nextSpec = StructSpec(self.spec[1][1])
            elif self.spec[0] in (TType.LIST, TType.SET):
                self.nextSpec = ListOrSetSpec(self.spec[1])
            elif self.spec[0] == TType.MAP:
                self.nextSpec = MapSpec(self.spec[1])
        return self.nextSpec


class MapSpec(ThriftSpec):
    '''Wraps a map's 4-tuple key/vale type spec.
    '''
    def __init__(self, spec):
        ThriftSpec.__init__(self, spec)
        self.key = True

        self.keySpec = None
        if self.spec[1] is not None:
            if self.spec[0] == TType.STRUCT:
                self.keySpec = StructSpec(self.spec[1][1])
            elif self.spec[0] in (TType.LIST, TType.SET):
                self.keySpec = ListOrSetSpec(self.spec[1])
            elif self.spec[0] == TType.MAP:
                self.keySpec = MapSpec(self.spec[1])

        self.valueSpec = None
        if self.spec[3] is not None:
            if self.spec[2] == TType.STRUCT:
                self.valueSpec = StructSpec(self.spec[3][1])
            elif self.spec[2] in (TType.LIST, TType.SET):
                self.valueSpec = ListOrSetSpec(self.spec[3])
            elif self.spec[2] == TType.MAP:
                self.valueSpec = MapSpec(self.spec[3])

    def readMapBegin(self):
        self.getNextSpec()
        return (self.spec[0], self.spec[2], -1)

    def getNextSpec(self):
        if self.keySpec is not None and self.valueSpec is not None:
            self.nextSpec = self.keySpec if self.key is True else \
                    self.valueSpec
            self.key = not self.key
        else:
            self.nextSpec = self.keySpec if self.keySpec is not None else \
                    self.valueSpec

        return self.nextSpec


class TSimpleJSONProtocolBase(TProtocolBase, object):
    def __init__(self, trans, spec=None):
        TProtocolBase.__init__(self, trans)
        # Used as stack for contexts.
        self.contexts = [TJSONContext(protocol=self)]
        self.context = TJSONContext(protocol=self)
        self.pair_context_class = TJSONPairContext
        self.list_context_class = TJSONListContext
        self.reader = LookaheadReader(self)
        self.specs = []
        self.spec = StructSpec(spec)

    def pushContext(self, newContext):
        self.contexts.append(self.context)
        self.context = newContext

    def popContext(self):
        if len(self.contexts) > 0:
            self.context = self.contexts.pop()

    def pushSpec(self, newSpec):
        self.specs.append(self.spec)
        self.spec = newSpec

    def popSpec(self):
        if len(self.specs) > 0:
            self.spec = self.specs.pop()

    def skipWhitespace(self):
        skipped = 0
        while True:
            ch = self.reader.peek()
            if ch not in WHITESPACE_CHARS:
                break
            self.reader.read()
            skipped += 1
        return skipped

    def skip(self, _type):
        self.context.read(self.reader)
        self.skipWhitespace()
        type = self.guessTypeIdFromFirstByte()
        # Since self.context.read is called at the beginning of all readJSONxxx
        # methods and we have already called it here, push an empty context so that
        # it becomes a no-op.
        self.pushContext(TJSONContext(protocol=self))
        if type == TType.STRUCT:
            self.readJSONObjectStart()
            while True:
                (_, ftype, _) = self.readFieldBegin()
                if ftype == TType.STOP:
                    break
                self.skip(TType.VOID)
            self.readJSONObjectEnd()
        elif type == TType.LIST:
            self.readJSONArrayStart()
            while self.peekList():
                self.skip(TType.VOID)
            self.readJSONArrayEnd()
        elif type == TType.STRING:
            self.readJSONString()
        elif type == TType.DOUBLE:
            self.readJSONDouble()
        elif type == TType.BOOL:
            self.readJSONBool()
        else:
            raise TProtocolException(
                TProtocolException.INVALID_DATA,
                "Unexpected type {} guessed when skipping".format(type)
            )
        self.popContext()

    def guessTypeIdFromFirstByte(self):
        self.skipWhitespace()
        byte = self.reader.peek()
        if byte == JSON_OBJECT_END or byte == JSON_ARRAY_END:
            return TType.STOP
        elif byte == JSON_STRING_DELIMITER:
            return TType.STRING
        elif byte == JSON_OBJECT_START:
            return TType.STRUCT
        elif byte == JSON_ARRAY_START:
            return TType.LIST
        elif byte == b't' or byte == b'f':
            return TType.BOOL
        elif byte in (b'+', b'-', b'0', b'1', b'2', b'3', b'4', b'5',
                b'6', b'7', b'8', b'9'):
            return TType.DOUBLE
        else:
            raise TProtocolException(TProtocolException.INVALID_DATA,
                    "Unrecognized byte: {}".format(byte))

    def writeJSONEscapeChar(self, ch):
        self.trans.write(JSON_ESCAPE_PREFIX)
        self.trans.write(hexChar(ch >> 4))
        self.trans.write(hexChar(ch))

    def writeJSONChar(self, ch):
        charValue = ord(ch)
        if charValue >= 0x30:
            # The only special character >= 0x30 is '\'.
            if charValue == JSON_BACKSLASH_VALUE:
                self.trans.write(JSON_BACKSLASH)
                self.trans.write(JSON_BACKSLASH)
            else:
                self.trans.write(ch)
        else:
            outCh = JSON_CHAR_TABLE[charValue]
            if outCh == 1:
                self.trans.write(ch)
            elif outCh:
                self.trans.write(JSON_BACKSLASH)
                self.trans.write(outCh)
            else:
                self.writeJSONEscapeChar(charValue)

    def writeJSONString(self, outStr):
        self.context.write(self.trans)
        self.trans.write(JSON_STRING_DELIMITER)
        outStrLen = len(outStr)
        if outStrLen > 0:
            is_int = isinstance(outStr[0], int)
            pos = 0
            for idx, ch in enumerate(outStr):
                if is_int:
                    ch = outStr[idx:idx + 1]
                if ch in JSON_CHARS_TO_ESCAPE:
                    if pos < idx:
                        # Write previous chunk not requiring escaping
                        self.trans.write(outStr[pos:idx])
                    # Write current char with escaping
                    self.writeJSONChar(ch)
                    # Advance pos
                    pos = idx + 1
            if pos < outStrLen:
                # Write last chunk till outStrLen
                self.trans.write(outStr[pos:outStrLen])
        self.trans.write(JSON_STRING_DELIMITER)

    def writeJSONBase64(self, outStr):
        self.context.write(self.trans)
        self.trans.write(JSON_STRING_DELIMITER)
        b64Str = b64encode(outStr)
        self.trans.write(b64Str)
        self.trans.write(JSON_STRING_DELIMITER)

    def writeJSONInteger(self, num):
        self.context.write(self.trans)
        escapeNum = self.context.escapeNum()
        numStr = str(num)
        if escapeNum:
            self.trans.write(JSON_STRING_DELIMITER)
        self.trans.write(numStr)
        if escapeNum:
            self.trans.write(JSON_STRING_DELIMITER)

    def writeJSONBool(self, boolVal):
        self.context.write(self.trans)
        if self.context.escapeNum():
            self.trans.write(JSON_STRING_DELIMITER)
        if boolVal:
            self.trans.write(b"true")
        else:
            self.trans.write(b"false")
        if self.context.escapeNum():
            self.trans.write(JSON_STRING_DELIMITER)

    def writeJSONDouble(self, num):
        self.context.write(self.trans)
        numStr = str(num)
        special = False

        if numStr == "nan":
            numStr = THRIFT_NAN
            special = True
        elif numStr == "inf":
            numStr = THRIFT_INFINITY
            special = True
        elif numStr == "-inf":
            numStr = THRIFT_NEGATIVE_INFINITY
            special = True

        escapeNum = special or self.context.escapeNum()
        if escapeNum:
            self.trans.write(JSON_STRING_DELIMITER)
        self.trans.write(numStr)
        if escapeNum:
            self.trans.write(JSON_STRING_DELIMITER)

    def writeJSONObjectStart(self):
        self.context.write(self.trans)
        self.trans.write(JSON_OBJECT_START)
        self.pushContext(self.pair_context_class(
            protocol=self,
            indentLevel=len(self.contexts)))

    def writeJSONObjectEnd(self):
        self.popContext()
        self.context.writeNewLine(self.trans)
        self.trans.write(JSON_OBJECT_END)

    def writeJSONArrayStart(self):
        self.context.write(self.trans)
        self.trans.write(JSON_ARRAY_START)
        self.pushContext(self.list_context_class(
            protocol=self,
            indentLevel=len(self.contexts)))

    def writeJSONArrayEnd(self):
        self.popContext()
        self.context.writeNewLine(self.trans)
        self.trans.write(JSON_ARRAY_END)

    def writeJSONMapStart(self):
        self.context.write(self.trans)
        self.trans.write(JSON_OBJECT_START)
        self.pushContext(self.list_context_class(
            protocol=self,
            indentLevel=len(self.contexts)))

    def writeJSONMapEnd(self):
        self.popContext()
        self.context.writeNewLine(self.trans)
        self.trans.write(JSON_OBJECT_END)

    def readJSONSyntaxChar(self, char):
        ch = self.reader.read()
        if ch != char:
            raise TProtocolException(TProtocolException.INVALID_DATA,
                                     "Unexpected character: %s" % ch)

    def readJSONString(self, skipContext=False):
        self.skipWhitespace()
        if skipContext is False:
            self.context.read(self.reader)
        self.skipWhitespace()
        self.readJSONSyntaxChar(JSON_STRING_DELIMITER)
        string = []
        while True:
            ch = self.reader.read()
            if ch == JSON_STRING_DELIMITER:
                break
            if ch == JSON_BACKSLASH:
                ch = self.reader.read()
                if ch == b'u':
                    self.readJSONSyntaxChar(JSON_ZERO_CHAR)
                    self.readJSONSyntaxChar(JSON_ZERO_CHAR)
                    data = self.trans.read(2)
                    if sys.version_info[0] >= 3 and isinstance(data, bytes):
                        ch = json.JSONDecoder().decode(
                            '"\\u00%s"' % str(data, 'utf-8')).encode('utf-8')
                    else:
                        ch = json.JSONDecoder().decode('"\\u00%s"' % data)
                else:
                    idx = ESCAPE_CHARS.find(ch)
                    if idx == -1:
                        raise TProtocolException(
                                TProtocolException.INVALID_DATA,
                                "Expected control char")
                    ch = ESCAPE_CHAR_VALS[idx]
            string.append(ch)
        return b''.join(string)

    def isJSONNumeric(self, ch):
        return NUMERIC_CHAR.find(ch) >= 0

    def readJSONNumericChars(self):
        numeric = []
        while True:
            ch = self.reader.peek()
            if self.isJSONNumeric(ch) is False:
                break
            numeric.append(self.reader.read())
        return b''.join(numeric)

    def readJSONInteger(self):
        self.context.read(self.reader)
        self.skipWhitespace()
        if self.context.escapeNum():
            self.readJSONSyntaxChar(JSON_STRING_DELIMITER)
        numeric = self.readJSONNumericChars()
        if self.context.escapeNum():
            self.readJSONSyntaxChar(JSON_STRING_DELIMITER)
        try:
            return int(numeric)
        except ValueError:
            raise TProtocolException(TProtocolException.INVALID_DATA,
                                     "Bad data encounted in numeric data")

    def readJSONDouble(self):
        self.context.read(self.reader)
        self.skipWhitespace()
        if self.reader.peek() == JSON_STRING_DELIMITER:
            string = self.readJSONString(True)
            try:
                double = float(string)
                if (self.context.escapeNum is False and
                    double != float('inf') and
                    double != float('-inf') and
                    double != float('nan')
                ):
                    raise TProtocolException(TProtocolException.INVALID_DATA,
                            "Numeric data unexpectedly quoted")
                return double
            except ValueError:
                raise TProtocolException(TProtocolException.INVALID_DATA,
                        "Bad data encountered in numeric data")
        else:
            try:
                return float(self.readJSONNumericChars())
            except ValueError:
                raise TProtocolException(TProtocolException.INVALID_DATA,
                        "Bad data encountered in numeric data")

    def readJSONBase64(self):
        string = self.readJSONString()
        return b64decode(string)

    def readJSONBool(self):
        self.context.read(self.reader)
        self.skipWhitespace()
        if self.context.escapeNum():
            self.readJSONSyntaxChar(JSON_STRING_DELIMITER)
        if self.reader.peek() == b't':
            true_string = b'true'
            for i in range(4):
                if self.reader.read() != true_string[i:i+1]:
                    raise TProtocolException(TProtocolException.INVALID_DATA,
                            "Bad data encountered in bool")
            boolVal = True
        elif self.reader.peek() == b'f':
            false_string = b'false'
            for i in range(5):
                if self.reader.read() != false_string[i:i+1]:
                    raise TProtocolException(TProtocolException.INVALID_DATA,
                            "Bad data encountered in bool")
            boolVal = False
        else:
            raise TProtocolException(TProtocolException.INVALID_DATA,
                    "Bad data encountered in bool")
        if self.context.escapeNum():
            self.readJSONSyntaxChar(JSON_STRING_DELIMITER)
        return boolVal

    def readJSONArrayStart(self):
        self.context.read(self.reader)
        self.skipWhitespace()
        self.readJSONSyntaxChar(JSON_ARRAY_START)
        self.pushContext(self.list_context_class(
            protocol=self,
            indentLevel=len(self.contexts)))

    def readJSONArrayEnd(self):
        self.popContext()
        self.skipWhitespace()
        self.readJSONSyntaxChar(JSON_ARRAY_END)

    def readJSONMapStart(self):
        self.context.read(self.reader)
        self.skipWhitespace()
        self.readJSONSyntaxChar(JSON_OBJECT_START)
        self.pushContext(self.list_context_class(
            protocol=self,
            indentLevel=len(self.contexts)))

    def readJSONMapEnd(self):
        self.popContext()
        self.skipWhitespace()
        self.readJSONSyntaxChar(JSON_OBJECT_END)

    def readJSONObjectStart(self):
        self.context.read(self.reader)
        self.skipWhitespace()
        self.readJSONSyntaxChar(JSON_OBJECT_START)
        self.pushContext(self.pair_context_class(
            protocol=self,
            indentLevel=len(self.contexts)))

    def readJSONObjectEnd(self):
        self.popContext()
        self.skipWhitespace()
        self.readJSONSyntaxChar(JSON_OBJECT_END)


class TSimpleJSONProtocol(TSimpleJSONProtocolBase):
    """
    JSON protocol implementation for Thrift. This protocol is write-only, and
    produces a simple output format that conforms to the JSON standard.
    """

    def writeMessageBegin(self, name, messageType, seqId):
        self.writeJSONArrayStart()
        self.context.writeNewLine(self.trans)
        self.writeJSONInteger(THRIFT_VERSION_1)
        self.writeJSONString(name)
        self.writeJSONInteger(messageType)
        self.writeJSONInteger(seqId)

    def writeMessageEnd(self):
        self.writeJSONArrayEnd()

    def writeStructBegin(self, name):
        self.writeJSONObjectStart()

    def writeStructEnd(self):
        self.writeJSONObjectEnd()

    def writeFieldBegin(self, name, fieldType, fieldId):
        self.context.write(self.trans)
        self.popContext()
        self.pushContext(self.pair_context_class(
            protocol=self,
            indentLevel=len(self.contexts)))
        self.context.writeNewLine(self.trans)
        self.writeJSONString(name)

    def writeFieldEnd(self):
        return

    def writeFieldStop(self):
        return

    def writeMapBegin(self, keyType, valType, size):
        self.writeJSONMapStart()
        self.context.writeNewLine(self.trans)
        self.pushContext(self.pair_context_class(
            protocol=self,
            indentLevel=len(self.contexts) - 1, isMapPair=True))

    def writeMapEnd(self):
        self.popContext()
        self.writeJSONMapEnd()

    def writeListBegin(self, elemType, size):
        self.writeJSONArrayStart()
        self.context.writeNewLine(self.trans)

    def writeListEnd(self):
        self.writeJSONArrayEnd()

    def writeSetBegin(self, elemType, size):
        self.writeJSONArrayStart()
        self.context.writeNewLine(self.trans)

    def writeSetEnd(self):
        self.writeJSONArrayEnd()

    def writeBool(self, val):
        self.writeJSONBool(val)

    def writeByte(self, byte):
        self.writeJSONInteger(byte)

    def writeI16(self, i16):
        self.writeJSONInteger(i16)

    def writeI32(self, i32):
        self.writeJSONInteger(i32)

    def writeI64(self, i64):
        self.writeJSONInteger(i64)

    def writeDouble(self, d):
        self.writeJSONDouble(d)

    def writeFloat(self, f):
        self.writeJSONDouble(f)

    def writeString(self, outStr):
        self.writeJSONString(outStr)

    def writeBinary(self, outStr):
        self.writeJSONBase64(outStr)

    def readMessageBegin(self):
        self.readJSONArrayStart()
        self.skipWhitespace()
        if self.readJSONInteger() != THRIFT_VERSION_1:
            raise TProtocolException(TProtocolException.BAD_VERSION,
                                     "Message contained bad version.")
        name = self.readJSONString()
        mtype = self.readJSONInteger()
        seqid = self.readJSONInteger()

        return (name, mtype, seqid)

    def readMessageEnd(self):
        self.readJSONArrayEnd()

    def readStructBegin(self):
        self.readJSONObjectStart()
        # This is needed because of the very first call
        if self.spec.nextSpec is not None:
            self.pushSpec(self.spec.getNextSpec())

    def readStructEnd(self):
        self.readJSONObjectEnd()
        self.popSpec()

    def readFieldBegin(self):
        self.skipWhitespace()
        ch = self.reader.peek()
        if ch == JSON_OBJECT_END:
            return (None, TType.STOP, 0)
        self.context.read(self.reader)
        self.popContext()
        self.pushContext(self.pair_context_class(
            protocol=self,
            indentLevel=len(self.contexts)))
        self.skipWhitespace()
        fname = self.readJSONString()
        self.skipWhitespace()
        self.readJSONSyntaxChar(JSON_PAIR_SEPARATOR)
        self.context.skipColon = True
        self.skipWhitespace()
        if self.reader.peek() == b'n':
            for i in range(4):
                if self.reader.read() != b'null'[i:i + 1]:
                    raise TProtocolException(
                        TProtocolException.INVALID_DATA,
                        "Bad data encountered in null",
                    )

            self.context.read(self.reader)  # "consume" the colon we skipped
            return self.readFieldBegin()

        assert isinstance(self.spec, StructSpec)
        return self.spec.readFieldBegin(
                fname,
                self.guessTypeIdFromFirstByte)

    def readFieldEnd(self):
        return

    def readFieldStop(self):
        return

    def readNumber(self):
        return self.readJSONInteger()
    readByte = readNumber
    readI16 = readNumber
    readI32 = readNumber
    readI64 = readNumber

    def readDouble(self):
        return self.readJSONDouble()

    def readFloat(self):
        return self.readJSONDouble()

    def readString(self):
        return self.readJSONString()

    def readBinary(self):
        return self.readJSONBase64()

    def readBool(self):
        return self.readJSONBool()

    def readMapBegin(self):
        self.readJSONMapStart()
        self.skipWhitespace()
        self.pushContext(self.pair_context_class(
            protocol=self,
            indentLevel=len(self.contexts) - 1, isMapPair=True))
        self.pushSpec(self.spec.getNextSpec())
        return self.spec.readMapBegin()

    def readMapEnd(self):
        self.popContext()
        self.readJSONMapEnd()
        self.popSpec()

    def peekMap(self):
        self.skipWhitespace()
        return self.reader.peek() != JSON_OBJECT_END

    def peekList(self):
        self.skipWhitespace()
        return self.reader.peek() != JSON_ARRAY_END
    peekSet = peekList

    def readListBegin(self):
        self.skipWhitespace()
        self.readJSONArrayStart()
        self.pushSpec(self.spec.getNextSpec())
        return self.spec.readListBegin()
    readSetBegin = readListBegin

    def readListEnd(self):
        self.skipWhitespace()
        self.readJSONArrayEnd()
        self.popSpec()
    readSetEnd = readListEnd


class TSimpleJSONProtocolFactory:
    def getProtocol(self, trans, spec=None):
        prot = TSimpleJSONProtocol(trans, spec)
        return prot



# --coding:utf-8--
#
# Copyright (c) 2021 vesoft inc. All rights reserved.
#
# This source code is licensed under Apache 2.0 License,
# attached with Common Clause Condition 1.0, found in the LICENSES directory.

import contextlib
import logging
import socket

from collections import deque
from threading import RLock, Timer

from nebula2.Exception import NotValidConnectionException, InValidHostname

from nebula2.gclient.net.Session import Session
from nebula2.gclient.net.Connection import Connection


class ConnectionPool(object):
    S_OK = 0
    S_BAD = 1

    def __init__(self):
        # all addresses of servers
        self._addresses = list()

        # server's status
        self._addresses_status = dict()

        # all connections
        self._connections = dict()
        self._configs = None
        self._ssl_configs = None
        self._lock = RLock()
        self._pos = -1
        self._close = False

    def __del__(self):
        self.close()

    def init(self, addresses, configs, ssl_conf=None):
        """init the connection pool

        :param addresses: the graphd servers' addresses
        :param configs: the config of the pool
        :param ssl_conf: the config of SSL socket
        :return: if all addresses are ok, return True else return False.
        """
        if self._close:
            logging.error('The pool has init or closed.')
            raise RuntimeError('The pool has init or closed.')
        self._configs = configs
        for address in addresses:
            if address not in self._addresses:
                try:
                    ip = socket.gethostbyname(address[0])
                except Exception:
                    raise InValidHostname(str(address[0]))
                ip_port = (ip, address[1])
                self._addresses.append(ip_port)
                self._addresses_status[ip_port] = self.S_BAD
                self._connections[ip_port] = deque()
        self._ssl_configs = ssl_conf
        self.update_servers_status()

        # detect the services
        self._period_detect()

        # init min connections
        ok_num = self.get_ok_servers_num()
        if ok_num < len(self._addresses):
            raise RuntimeError(
                'The services status exception: {}'.format(self._get_services_status())
            )

        conns_per_address = int(self._configs.min_connection_pool_size / ok_num)

        for addr in self._addresses:
            for i in range(0, conns_per_address):
                connection = Connection()
                connection.open_SSL(
                    addr[0], addr[1], self._configs.timeout, self._ssl_configs
                )
                self._connections[addr].append(connection)
        return True

    def get_session(self, user_name, password, retry_connect=True):
        """get session

        :param user_name: the user name to authenticate graphd
        :param password: the password to authenticate graphd
        :param retry_connect:
        :return: Session
        """
        connection = self.get_connection()
        if connection is None:
            raise NotValidConnectionException()
        try:
            auth_result = connection.authenticate(user_name, password)
            return Session(connection, auth_result, self, retry_connect)
        except Exception:
            raise

    @contextlib.contextmanager
    def session_context(self, *args, **kwargs):
        """
        session_context is to be used with a contextlib.contextmanager.
        It returns a connection session from the pool, with same params
        as the method get_session().

        When session_context is exited, the connection will be released.

        :param user_name: the user name to authenticate graphd
        :param password: the password to authenticate graphd
        :param retry_connect: if auto retry connect
        :return: contextlib._GeneratorContextManager
        """
        session = None
        try:
            session = self.get_session(*args, **kwargs)
            yield session
        except Exception:
            raise
        finally:
            if session:
                session.release()

    def get_connection(self):
        """get available connection

        :return: Connection
        """
        with self._lock:
            if self._close:
                logging.error('The pool is closed')
                raise NotValidConnectionException()

            try:
                ok_num = self.get_ok_servers_num()
                if ok_num == 0:
                    logging.error('No available server')
                    return None
                max_con_per_address = int(
                    self._configs.max_connection_pool_size / ok_num
                )
                try_count = 0
                while try_count <= len(self._addresses):
                    self._pos = (self._pos + 1) % len(self._addresses)
                    addr = self._addresses[self._pos]
                    if self._addresses_status[addr] == self.S_OK:
                        invalid_connections = list()

                        # iterate all connections to find an available connection
                        for connection in self._connections[addr]:
                            if not connection.is_used:
                                # ping to check the connection is valid
                                if connection.ping():
                                    connection.is_used = True
                                    logging.info('Get connection to {}'.format(addr))
                                    return connection
                                else:
                                    invalid_connections.append(connection)

                        # remove invalid connections
                        for connection in invalid_connections:
                            self._connections[addr].remove(connection)

                        # check if the server is still alive
                        if not self.ping(addr):
                            self._addresses_status[addr] = self.S_BAD
                            continue

                        # create new connection if the number of connections is less than max_con_per_address
                        if len(self._connections[addr]) < max_con_per_address:
                            connection = Connection()
                            connection.open_SSL(
                                addr[0],
                                addr[1],
                                self._configs.timeout,
                                self._ssl_configs,
                            )
                            connection.is_used = True
                            self._connections[addr].append(connection)
                            logging.info('Get connection to {}'.format(addr))
                            return connection
                    else:
                        for connection in list(self._connections[addr]):
                            if not connection.is_used:
                                self._connections[addr].remove(connection)
                    try_count = try_count + 1

                logging.error('No available connection')
                return None
            except Exception as ex:
                logging.error('Get connection failed: {}'.format(ex))
                return None

    def ping(self, address):
        """check the server is ok

        :param address: the server address want to connect
        :return: True or False
        """
        try:
            conn = Connection()
            conn.open_SSL(address[0], address[1], 1000, self._ssl_configs)
            conn.close()
            return True
        except Exception as ex:
            logging.warning(
                'Connect {}:{} failed: {}'.format(address[0], address[1], ex)
            )
            return False

    def close(self):
        """close all connections in pool

        :return: void
        """
        with self._lock:
            for addr in self._connections.keys():
                for connection in self._connections[addr]:
                    if connection.is_used:
                        logging.warning('Closing a connection that is in use')
                    connection.close()
            self._close = True

    def connects(self):
        """get the number of existing connections

        :return: the number of connections
        """
        with self._lock:
            count = 0
            for addr in self._connections.keys():
                count = count + len(self._connections[addr])
            return count

    def in_used_connects(self):
        """get the number of the used connections

        :return: int
        """
        with self._lock:
            count = 0
            for addr in self._connections.keys():
                for connection in self._connections[addr]:
                    if connection.is_used:
                        count = count + 1
            return count

    def get_ok_servers_num(self):
        """get the number of the ok servers

        :return: int
        """
        count = 0
        for addr in self._addresses_status.keys():
            if self._addresses_status[addr] == self.S_OK:
                count = count + 1
        return count

    def _get_services_status(self):
        msg_list = []
        for addr in self._addresses_status.keys():
            status = 'OK'
            if self._addresses_status[addr] != self.S_OK:
                status = 'BAD'
            msg_list.append('[services: {}, status: {}]'.format(addr, status))
        return ', '.join(msg_list)

    def update_servers_status(self):
        """update the servers' status"""
        for address in self._addresses:
            if self.ping(address):
                self._addresses_status[address] = self.S_OK
            else:
                self._addresses_status[address] = self.S_BAD

    def _remove_idle_unusable_connection(self):
        if self._configs.idle_time == 0:
            return
        with self._lock:
            for addr in self._connections.keys():
                conns = self._connections[addr]
                for connection in list(conns):
                    if not connection.is_used:
                        if not connection.ping():
                            logging.debug(
                                'Remove the unusable connection to {}'.format(
                                    connection.get_address()
                                )
                            )
                            conns.remove(connection)
                            continue
                        if (
                            self._configs.idle_time != 0
                            and connection.idle_time() > self._configs.idle_time
                        ):
                            logging.debug(
                                'Remove the idle connection to {}'.format(
                                    connection.get_address()
                                )
                            )
                            conns.remove(connection)

    def _period_detect(self):
        if self._close or self._configs.interval_check < 0:
            return
        self.update_servers_status()
        self._remove_idle_unusable_connection()
        timer = Timer(self._configs.interval_check, self._period_detect)
        timer.setDaemon(True)
        timer.start()



# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from os import path
from SCons.Builder import Builder

def scons_env(env, add=''):
    opath = path.dirname(path.abspath('$TARGET'))
    lstr = 'thrift --gen cpp -o ' + opath + ' ' + add + ' $SOURCE'
    cppbuild = Builder(action=lstr)
    env.Append(BUILDERS={'ThriftCpp': cppbuild})

def gen_cpp(env, dir, file):
    scons_env(env)
    suffixes = ['_data.h', '_data.cpp', '_types.h', '_types.cpp']
    targets = ['gen-cpp' + file + s for s in suffixes]
    return env.ThriftCpp(targets, dir + file + '.thrift')



#!/usr/bin/env python
# --coding:utf-8--

# Copyright (c) 2020 vesoft inc. All rights reserved.
#
# This source code is licensed under Apache 2.0 License,
# attached with Common Clause Condition 1.0, found in the LICENSES directory.

import pytz
from datetime import datetime, timezone, timedelta
from nebula2.Exception import (
    InvalidValueTypeException,
    InvalidKeyException,
    OutOfRangeException
)

from nebula2.common.ttypes import Geography, Value, Vertex, Edge, NullType, DateTime, Time


def date_time_convert_with_timezone(date_time: DateTime, timezone_offset: int):
    """the function to convert utc date_time to local date_time

    :param date_time: the utc date_time
    :param timezone_offset: the timezone offset
    :return: the date_time with timezone
    """
    native_date_time = datetime(date_time.year,
                                date_time.month,
                                date_time.day,
                                date_time.hour,
                                date_time.minute,
                                date_time.sec,
                                date_time.microsec,
                                pytz.timezone("utc"))
    local_date_time = native_date_time.astimezone(
        timezone(timedelta(seconds=timezone_offset)))
    new_date_time = DateTime()
    new_date_time.year = local_date_time.year
    new_date_time.month = local_date_time.month
    new_date_time.day = local_date_time.day
    new_date_time.hour = local_date_time.hour
    new_date_time.minute = local_date_time.minute
    new_date_time.sec = local_date_time.second
    new_date_time.microsec = local_date_time.microsecond
    return new_date_time


def time_convert_with_timezone(n_time: Time, timezone_offset: int):
    """the function to convert utc date_time to local date_time

    :param n_time: the utc time
    :param timezone_offset: the timezone offset
    :return: the time with the timezone
    """
    native_date_time = datetime(1,
                                1,
                                1,
                                n_time.hour,
                                n_time.minute,
                                n_time.sec,
                                n_time.microsec,
                                pytz.timezone("utc"))
    local_date_time = native_date_time.astimezone(
        timezone(timedelta(seconds=timezone_offset)))
    local_time = Time()
    local_time.hour = local_date_time.hour
    local_time.minute = local_date_time.minute
    local_time.sec = local_date_time.second
    local_time.microsec = local_date_time.microsecond
    return local_time


class BaseObject(object):
    def __init__(self):
        self._decode_type = 'utf-8'
        self._timezone_offset = 0

    def set_decode_type(self, decode_type):
        self._decode_type = decode_type
        return self

    def set_timezone_offset(self, timezone_offset):
        self._timezone_offset = timezone_offset
        return self

    def get_decode_type(self):
        return self._decode_type

    def get_timezone_offset(self):
        return self._timezone_offset


class Record(object):
    def __init__(self, values, names, decode_type='utf-8', timezone_offset: int = 0):
        assert len(names) == len(values),\
            'len(names): {} != len(values): {}, names: {}, values: {}'.format(
                len(names), len(values), str(names), str(values))
        self._record = list()
        self._names = names

        for val in values:
            self._record.append(ValueWrapper(val,
                                             decode_type=decode_type,
                                             timezone_offset=timezone_offset))

    def __iter__(self):
        return iter(self._record)

    def size(self):
        """the size of record

        :return: record size
        """
        return len(self._names)

    def get_value(self, index):
        """get value by specified index

        :param index: the index of column
        :return: ValueWrapper
        """
        if index >= len(self._names):
            raise OutOfRangeException()
        return self._record[index]

    def get_value_by_key(self, key):
        """get value by key

        :return: Value
        """
        try:
            return self._record[self._names.index(key)]
        except Exception:
            raise InvalidKeyException(key)

    def keys(self):
        """get column names of record

        :return: the column names
        """
        return self._names

    def values(self):
        """get all values

        :return: values
        """
        return self._record

    def __repr__(self):
        return "{}".format('\n'.join([str(val_wrap) for val_wrap in self._record]))

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class DataSetWrapper(object):
    def __init__(self, data_set, decode_type='utf-8', timezone_offset: int = 0):
        assert data_set is not None
        self._decode_type = decode_type
        self._timezone_offset = timezone_offset
        self._data_set = data_set
        self._column_names = []
        self._key_indexes = {}
        self._pos = -1
        for index, name in enumerate(self._data_set.column_names):
            d_name = name.decode(self._decode_type)
            self._column_names.append(d_name)
            self._key_indexes[d_name] = index

    def get_row_size(self):
        return len(self._data_set.rows)

    def get_col_names(self):
        return self._column_names

    def get_rows(self):
        return self._data_set.rows

    def get_row_types(self):
        """Get row types

        :param empty
        :return: list<int>
          ttypes.Value.__EMPTY__ = 0
          ttypes.Value.NVAL = 1
          ttypes.Value.BVAL = 2
          ttypes.Value.IVAL = 3
          ttypes.Value.FVAL = 4
          ttypes.Value.SVAL = 5
          ttypes.Value.DVAL = 6
          ttypes.Value.TVAL = 7
          ttypes.Value.DTVAL = 8
          ttypes.Value.VVAL = 9
          ttypes.Value.EVAL = 10
          ttypes.Value.PVAL = 11
          ttypes.Value.LVAL = 12
          ttypes.Value.MVAL = 13
          ttypes.Value.UVAL = 14
          ttypes.Value.GVAL = 15
          ttypes.Value.GGVAL = 16
        """
        if len(self._data_set.rows) == 0:
            return []
        return [(value.getType()) for value in self._data_set.rows[0].values]

    def row_values(self, row_index):
        """get row values

        :param row_index: the Record index
        :return: list<ValueWrapper>
        """
        if row_index >= len(self._data_set.rows):
            raise OutOfRangeException()
        return [(ValueWrapper(value=value,
                              decode_type=self._decode_type,
                              timezone_offset=self._timezone_offset))
                for value in self._data_set.rows[row_index].values]

    def column_values(self, key):
        """get column values

        :param key: the col name
        :return: list<ValueWrapper>
        """
        if key not in self._column_names:
            raise InvalidKeyException(key)

        return [(ValueWrapper(value=row.values[self._key_indexes[key]],
                              decode_type=self._decode_type,
                              timezone_offset=self._timezone_offset))
                for row in self._data_set.rows]

    def __iter__(self):
        self._pos = -1
        return self

    def __next__(self):
        """The record iterator

        :return: record
        """
        if len(self._data_set.rows) == 0 or self._pos >= len(self._data_set.rows) - 1:
            raise StopIteration
        self._pos = self._pos + 1
        return Record(values=self._data_set.rows[self._pos].values,
                      names=self._column_names,
                      decode_type=self._decode_type,
                      timezone_offset=self._timezone_offset)

    def __repr__(self):
        data_str = []
        for i in range(self.get_row_size()):
            data_str.append(str(self.row_values(i)))
        value_str = ','.join(data_str)
        return 'keys: {}, values: {}'.format(self._column_names, value_str)


class Null(object):
    __NULL__ = NullType.__NULL__
    NaN = NullType.NaN
    BAD_DATA = NullType.BAD_DATA
    BAD_TYPE = NullType.BAD_TYPE
    ERR_OVERFLOW = NullType.ERR_OVERFLOW
    UNKNOWN_PROP = NullType.UNKNOWN_PROP
    DIV_BY_ZERO = NullType.DIV_BY_ZERO
    OUT_OF_RANGE = NullType.OUT_OF_RANGE

    def __init__(self, type):
        self._type = type

    def __repr__(self):
        return NullType._VALUES_TO_NAMES[self._type]

    def __eq__(self, other):
        return self._type == other._type


class ValueWrapper(object):
    def __init__(self, value, decode_type='utf-8', timezone_offset: int = 0):
        self._value = value
        self._decode_type = decode_type
        self._timezone_offset = timezone_offset

    def get_value(self):
        """get raw data

        :return: Value
        """
        return self._value

    def is_null(self):
        """judge the value if is Null type

        :return: true or false
        """
        return self._value.getType() == Value.NVAL

    def is_empty(self):
        """judge the value if is Empty type

        :return: true or false
        """
        return self._value.getType() == Value.__EMPTY__

    def is_bool(self):
        """judge the value if is Bool type

        :return: true or false
        """
        return self._value.getType() == Value.BVAL

    def is_int(self):
        """judge the value if is Int type

        :return: true or false
        """
        return self._value.getType() == Value.IVAL

    def is_double(self):
        """judge the value if is Double type

        :return: true or false
        """
        return self._value.getType() == Value.FVAL

    def is_string(self):
        """judge the value if is String type

        :return: true or false
        """
        return self._value.getType() == Value.SVAL

    def is_list(self):
        """judge the value if is List type

        :return: true or false
        """
        return self._value.getType() == Value.LVAL

    def is_set(self):
        """judge the value if is Set type

        :return: true or false
        """
        return self._value.getType() == Value.UVAL

    def is_map(self):
        """judge the value if is Map type

        :return: true or false
        """
        return self._value.getType() == Value.MVAL

    def is_time(self):
        """judge the value if is Time type

        :return: true or false
        """
        return self._value.getType() == Value.TVAL

    def is_date(self):
        """judge the value if is Date type

        :return: true or false
        """
        return self._value.getType() == Value.DVAL

    def is_datetime(self):
        """judge the value if is Datetime type

        :return: true or false
        """
        return self._value.getType() == Value.DTVAL

    def is_vertex(self):
        """judge the value if is Vertex type

        :return: true or false
        """
        return self._value.getType() == Value.VVAL

    def is_edge(self):
        """judge the value if is Edge type

        :return: true or false
        """
        return self._value.getType() == Value.EVAL

    def is_path(self):
        """judge the value if is Path type

        :return: true or false
        """
        return self._value.getType() == Value.PVAL

    def is_geography(self):
        """judge the value if is Geography type

        :return: true or false
        """
        return self._value.getType() == Value.GGVAL

    def as_null(self):
        """converts the original data type to Null type

        :return: Null value
        """
        if self._value.getType() == Value.NVAL:
            return Null(self._value.get_nVal())
        raise InvalidValueTypeException(
            "expect NULL type, but is " + self._get_type_name())

    def as_bool(self):
        """converts the original data type to Bool type

        :return: Bool value
        """
        if self._value.getType() == Value.BVAL:
            return self._value.get_bVal()
        raise InvalidValueTypeException(
            "expect bool type, but is " + self._get_type_name())

    def as_int(self):
        """converts the original data type to Int type

        :return: Int value
        """
        if self._value.getType() == Value.IVAL:
            return self._value.get_iVal()
        raise InvalidValueTypeException(
            "expect bool type, but is " + self._get_type_name())

    def as_double(self):
        """converts the original data type to Double type

        :return: Double value
        """
        if self._value.getType() == Value.FVAL:
            return self._value.get_fVal()
        raise InvalidValueTypeException(
            "expect int type, but is " + self._get_type_name())

    def as_string(self):
        """converts the original data type to String type

        :return: String value
        """
        if self._value.getType() == Value.SVAL:
            return self._value.get_sVal().decode(self._decode_type)
        raise InvalidValueTypeException(
            "expect string type, but is " + self._get_type_name())

    def as_time(self):
        """converts the original data type to Time type

        :return: Time value
        """
        if self._value.getType() == Value.TVAL:
            return TimeWrapper(self._value.get_tVal()).set_timezone_offset(self._timezone_offset)
        raise InvalidValueTypeException(
            "expect time type, but is " + self._get_type_name())

    def as_date(self):
        """converts the original data type to Date type

        :return: Date value
        """
        if self._value.getType() == Value.DVAL:
            return DateWrapper(self._value.get_dVal())
        raise InvalidValueTypeException(
            "expect date type, but is " + self._get_type_name())

    def as_datetime(self):
        """converts the original data type to Datetime type

        :return: Datetime value
        """
        if self._value.getType() == Value.DTVAL:
            return DateTimeWrapper(self._value.get_dtVal()).set_timezone_offset(self._timezone_offset)
        raise InvalidValueTypeException(
            "expect datetime type, but is " + self._get_type_name())

    def as_list(self):
        """converts the original data type to list of ValueWrapper

        :return: list<ValueWrapper>
        """
        if self._value.getType() == Value.LVAL:
            result = []
            for val in self._value.get_lVal().values:
                result.append(ValueWrapper(val,
                                           decode_type=self._decode_type,
                                           timezone_offset=self._timezone_offset))
            return result
        raise InvalidValueTypeException(
            "expect list type, but is " + self._get_type_name())

    def as_set(self):
        """converts the original data type to set of ValueWrapper

        :return: set<ValueWrapper>
        """
        if self._value.getType() == Value.UVAL:
            result = set()
            for val in self._value.get_uVal().values:
                result.add(ValueWrapper(val,
                                        decode_type=self._decode_type,
                                        timezone_offset=self._timezone_offset))
            return result
        raise InvalidValueTypeException(
            "expect set type, but is " + self._get_type_name())

    def as_map(self):
        """converts the original data type to map type

        :return: map<String, ValueWrapper>
        """
        if self._value.getType() == Value.MVAL:
            result = {}
            kvs = self._value.get_mVal().kvs
            for key in kvs.keys():
                result[key.decode(self._decode_type)] = ValueWrapper(kvs[key],
                                                                     decode_type=self._decode_type,
                                                                     timezone_offset=self._timezone_offset)
            return result
        raise InvalidValueTypeException(
            "expect map type, but is " + self._get_type_name())

    def as_node(self):
        """converts the original data type to Node type

        :return: Node type
        """
        if self._value.getType() == Value.VVAL:
            return Node(self._value.get_vVal())\
                .set_decode_type(self._decode_type)\
                .set_timezone_offset(self._timezone_offset)
        raise InvalidValueTypeException(
            "expect vertex type, but is " + self._get_type_name())

    def as_relationship(self):
        """converts the original data type to Relationship type

        :return: Relationship type
        """
        if self._value.getType() == Value.EVAL:
            return Relationship(self._value.get_eVal())\
                .set_decode_type(self._decode_type)\
                .set_timezone_offset(self._timezone_offset)
        raise InvalidValueTypeException(
            "expect edge type, but is " + self._get_type_name())

    def as_path(self):
        """converts the original data type to PathWrapper type

        :return: PathWrapper type
        """
        if self._value.getType() == Value.PVAL:
            return PathWrapper(self._value.get_pVal())\
                .set_decode_type(self._decode_type)\
                .set_timezone_offset(self._timezone_offset)
        raise InvalidValueTypeException(
            "expect path type, but is " + self._get_type_name())

    def as_geography(self):
        """converts the original data type to GeographyWrapper type

        :return: GeographyWrapper type
        """
        if self._value.getType() == Value.GGVAL:
            return GeographyWrapper(self._value.get_ggVal())\
                .set_decode_type(self._decode_type)\
                .set_timezone_offset(self._timezone_offset)
        raise InvalidValueTypeException(
            "expect geography type, but is " + self._get_type_name())

    def _get_type_name(self):
        if self.is_empty():
            return "empty"
        if self.is_null():
            return "null"
        if self.is_bool():
            return "bool"
        if self.is_int():
            return "int"
        if self.is_double():
            return "double"
        if self.is_string():
            return "string"
        if self.is_list():
            return "list"
        if self.is_set():
            return "set"
        if self.is_map():
            return "map"
        if self.is_time():
            return "time"
        if self.is_date():
            return "date"
        if self.is_datetime():
            return "datetime"
        if self.is_vertex():
            return "vertex"
        if self.is_edge():
            return "edge"
        if self.is_path():
            return "path"
        if self.is_geography():
            return "geography"
        return "unknown"

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, self.__class__):
            return False
        if self.get_value().getType() != o.get_value().getType():
            return False
        if self.is_empty():
            return o.is_empty()
        elif self.is_null():
            return self.as_null() == o.as_null()
        elif self.is_bool():
            return self.as_bool() == o.as_bool()
        elif self.is_int():
            return self.as_int() == o.as_int()
        elif self.is_double():
            return self.as_double() == o.as_double()
        elif self.is_string():
            return self.as_string() == o.as_string()
        elif self.is_list():
            return self.as_list() == o.as_list()
        elif self.is_set():
            return self.as_set() == o.as_set()
        elif self.is_map():
            return self.as_map() == o.as_map()
        elif self.is_vertex():
            return self.as_node() == o.as_node()
        elif self.is_edge():
            return self.as_relationship() == o.as_relationship()
        elif self.is_path():
            return self.as_path() == o.as_path()
        elif self.is_time():
            return self.as_time() == self.as_time()
        elif self.is_date():
            return self.as_date() == self.as_date()
        elif self.is_datetime():
            return self.as_datetime() == self.as_datetime()
        elif self.is_geography():
            return self.as_geography() == self.as_geography()
        else:
            raise RuntimeError(
                'Unsupported type:{} to compare'.format(self._get_type_name()))
        return False

    def __repr__(self):
        if self.is_empty():
            return '__EMPTY__'
        elif self.is_null():
            return str(self.as_null())
        elif self.is_bool():
            return 'True' if self.as_bool() else 'False'
        elif self.is_int():
            return str(self.as_int())
        elif self.is_double():
            return str(self.as_double())
        elif self.is_string():
            return '\"{}\"'.format(self.as_string())
        elif self.is_list():
            return str(self.as_list())
        elif self.is_set():
            return str(self.as_set())
        elif self.is_map():
            return str(self.as_map())
        elif self.is_vertex():
            return str(self.as_node())
        elif self.is_edge():
            return str(self.as_relationship())
        elif self.is_path():
            return str(self.as_path())
        elif self.is_time():
            return str(self.as_time())
        elif self.is_date():
            return str(self.as_date())
        elif self.is_datetime():
            return str(self.as_datetime())
        elif self.is_geography():
            return str(self.as_geography())
        else:
            raise RuntimeError(
                'Unsupported type:{} to compare'.format(self._get_type_name()))
        return False

    def __hash__(self):
        return self._value.__hash__()


class TimeWrapper(BaseObject):
    def __init__(self, time):
        super(TimeWrapper, self).__init__()
        self._time = time

    def get_hour(self):
        """get utc hour

        :return: hour
        """
        return self._time.hour

    def get_minute(self):
        """get utc minute

        :return: minute
        """
        return self._time.minute

    def get_sec(self):
        """get utc second

        :return: second
        """
        return self._time.sec

    def get_microsec(self):
        """get utc microseconds

        :return: microseconds
        """
        return self._time.microsec

    def get_time(self):
        """get utc time

        :return: Time value
        """
        return self._time

    def get_local_time(self):
        """get time with the timezone from graph service

        :return: Time value with timezone offset
        """
        return time_convert_with_timezone(self._time, self.get_timezone_offset())

    def get_local_time_by_timezone_offset(self, timezone_offset):
        """get local time with the specified timezone by user

        :return: Time value with timezone offset
        """
        return time_convert_with_timezone(self._time, timezone_offset)

    def get_local_time_str(self):
        """convert local time string format

        :return: return local time string format
        """
        local_time = time_convert_with_timezone(
            self._time, self.get_timezone_offset())
        return "%02d:%02d:%02d.%06d" % (local_time.hour,
                                        local_time.minute,
                                        local_time.sec,
                                        local_time.microsec)

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return self._time.hour == other.get_hour() and \
            self._time.minute == other.get_minute() and \
            self._time.sec == other.get_sec() and \
            self._time.microsec == self.get_microsec()

    def __repr__(self):
        return "utc time: %02d:%02d:%02d.%06d, timezone_offset: %d" % (
            self._time.hour,
            self._time.minute,
            self._time.sec,
            self._time.microsec,
            self.get_timezone_offset())


class DateWrapper(object):
    def __init__(self, date):
        self._date = date

    def get_year(self):
        """get year

        :return: year
        """
        return self._date.year

    def get_month(self):
        """get month

        :return: month
        """
        return self._date.month

    def get_day(self):
        """get day

        :return: day
        """
        return self._date.day

    def get_date(self):
        """get original date

        :return: Date
        """
        return self._date

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return self._date.year == other.get_year() and \
            self._date.month == other.get_month() and \
            self._date.day == other.get_day()

    def __repr__(self):
        return "%d-%02d-%02d" % (self._date.year, self._date.month, self._date.day)


class DateTimeWrapper(BaseObject):
    def __init__(self, date_time):
        super(DateTimeWrapper, self).__init__()
        self._date_time = date_time

    def get_year(self):
        """get utc year

        :return: year
        """
        return self._date_time.year

    def get_month(self):
        """get utc month

        :return: month
        """
        return self._date_time.month

    def get_day(self):
        """get utc day

        :return: day
        """
        return self._date_time.day

    def get_hour(self):
        """get utc hour

        :return: hour
        """
        return self._date_time.hour

    def get_minute(self):
        """get utc minute

        :return: minute
        """
        return self._date_time.minute

    def get_sec(self):
        """get utc seconds

        :return: seconds
        """
        return self._date_time.sec

    def get_microsec(self):
        """get utc microseconds

        :return: microseconds
        """
        return self._date_time.microsec

    def get_datetime(self):
        """get utc datetime

        :return: datetime
        """
        return self._date_time

    def get_local_datetime(self):
        """get datetime with the timezone from graph service

        :return: Datetime value with timezone offset
        """
        return date_time_convert_with_timezone(self._date_time, self.get_timezone_offset())

    def get_local_datetime_by_timezone_offset(self, timezone_offset):
        """get local datetime with the specified timezone by user

        :return: Time value with timezone offset
        """
        return date_time_convert_with_timezone(self._date_time, timezone_offset)

    def get_local_datetime_str(self):
        """convert local datetime string format

        :return: return local datetime string format
        """
        local_date_time = date_time_convert_with_timezone(
            self._date_time, self.get_timezone_offset())
        return "%d-%02d-%02dT%02d:%02d:%02d.%06d" % (local_date_time.year,
                                                     local_date_time.month,
                                                     local_date_time.day,
                                                     local_date_time.hour,
                                                     local_date_time.minute,
                                                     local_date_time.sec,
                                                     local_date_time.microsec)

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return self._date_time.year == other.get_year() and \
            self._date_time.month == other.get_month() and \
            self._date_time.day == other.get_day() and \
            self._date_time.hour == other.get_hour() and \
            self._date_time.minute == other.get_minute() and \
            self._date_time.sec == other.get_sec() and \
            self._date_time.microsec == other.get_microsec()

    def __repr__(self):
        return "utc datetime: %d-%02d-%02dT%02d:%02d:%02d.%06d, timezone_offset: %d" % (
            self._date_time.year,
            self._date_time.month,
            self._date_time.day,
            self._date_time.hour,
            self._date_time.minute,
            self._date_time.sec,
            self._date_time.microsec,
            self.get_timezone_offset())


class CoordinateWrapper(BaseObject):
    def __init__(self, x, y):
        super(CoordinateWrapper, self).__init__()
        self._x = x
        self._y = y

    def get_x(self):
        """get x

        :return: double
        """
        return self._x

    def get_y(self):
        """get y

        :return: double
        """
        return self._y

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return self._x == other.get_x() and \
            self._y == other.get_y()

    def __repr__(self):
        return "%f %f" % (
            self._x,
            self._y)


class PointWrapper(BaseObject):
    def __init__(self, coord):
        super(PointWrapper, self).__init__()
        self._coord = coord

    def get_coordinate(self):
        """get raw data

        :return: CoordinateWrapper
        """
        return self._coord

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return self._coord == other.get_coordinate()

    def __repr__(self):
        return "POINT(" + str(self._coord) + ")"


class LineStringWrapper(BaseObject):
    def __init__(self, coord_list):
        super(LineStringWrapper, self).__init__()
        self._coord_list = coord_list

    def get_coordinate_list(self):
        """get raw data

        :return: list of CoordinateWrapper
        """
        return self._coord_list

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return self._coord_list == other.get_coordinate_list()

    def __repr__(self):
        wkt = "LINESTRING("
        for i in range(len(self._coord_list)):
            coord = self._coord_list[i]
            wkt += str(coord)
            if i < len(self._coord_list) - 1:
                wkt += ", "

        wkt += ")"
        return wkt


class PolygonWrapper(BaseObject):
    def __init__(self, coord_list_list):
        super(PolygonWrapper, self).__init__()
        self._coord_list_list = coord_list_list

    def get_coordinate_list_list(self):
        """get raw data

        :return: list of list of CoordinateWrapper
        """
        return self._coord_list_list

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return self._coord_list_list == other.get_coordinate_list_list()

    def __repr__(self):
        wkt = "POLYGON("
        for i in range(len(self._coord_list_list)):
            coord_list = self._coord_list_list[i]
            wkt += "("
            for j in range(len(coord_list)):
                coord = coord_list[i]
                wkt += str(coord)
                if i < len(self._coord_list) - 1:
                    wkt += ", "
            wkt += ")"
            if i < len(self._coord_list_list) - 1:
                wkt += ", "

        wkt += ")"
        return wkt


class GeographyWrapper(BaseObject):
    def __init__(self, geography):
        assert isinstance(geography, Geography)
        super(GeographyWrapper, self).__init__()
        self._geography = geography

    def get_geography(self):
        """get raw data

        :return: Geography
        """
        return self._geography

    def is_point(self):
        """judge the geography if is Point type

        :return: true or false
        """
        return self._geography.getType() == Geography.PTVAL

    def is_linestring(self):
        """judge the geography if is LineString type

        :return: true or false
        """
        return self._geography.getType() == Geography.LSVAL

    def is_polygon(self):
        """judge the geography if is Polygon type

        :return: true or false
        """
        return self._geography.getType() == Geography.PGVAL

    def as_point(self):
        """converts the original data type to Point type

        :return: PointWrapper
        """
        if self._geography.getType() == Geography.PTVAL:
            return PointWrapper(self._geography.get_ptVal())
        raise InvalidValueTypeException(
            "expect Point type, but is " + self._get_type_name())

    def as_linestring(self):
        """converts the original data type to LineString type

        :return: LineStringWrapper
        """
        if self._geography.getType() == Geography.LSVAL:
            return LineStringWrapper(self._geography.get_lsVal())
        raise InvalidValueTypeException(
            "expect LineString type, but is " + self._get_type_name())

    def as_polygon(self):
        """converts the original data type to Polygon type

        :return: PolygonWrapper
        """
        if self._geography.getType() == Geography.PGVAL:
            return PolygonWrapper(self._geography.get_pgVal())
        raise InvalidValueTypeException(
            "expect Polygon type, but is " + self._get_type_name())

    def _get_type_name(self):
        if self.is_point():
            return "point"
        if self.is_linestring():
            return "linestring"
        if self.is_polygon():
            return "polygon"
        return "unknown"

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, self.__class__):
            return False
        if self.is_point():
            return self.as_point() == o.as_point()
        elif self.is_linestring():
            return self.as_linestring() == o.as_linestring()
        elif self.is_polygon():
            return self.as_polygon() == o.as_polygon()
        else:
            raise RuntimeError(
                'Unsupported type:{} to compare'.format(self._get_type_name()))

    def __repr__(self):
        if self.is_point():
            return str(self.as_point())
        elif self.is_linestring():
            return str(self.as_linestring())
        elif self.is_polygon():
            return str(self.as_polygon())
        else:
            raise RuntimeError(
                'Unsupported type:{} to compare'.format(self._get_type_name()))


class GenValue(object):
    @classmethod
    def gen_vertex(cls, vid, tags):
        vertex = Vertex()
        vertex.vid = vid
        vertex.tags = tags
        return vertex

    @classmethod
    def gen_edge(cls, src_id, dst_id, type, edge_name, ranking, props):
        edge = Edge()
        edge.src = src_id
        edge.dst = dst_id
        edge.type = type
        edge.name = edge_name
        edge.ranking = ranking
        edge.props = props
        return edge

    @classmethod
    def gen_segment(cls, start_node, end_node, relationship):
        segment = Segment()
        segment.start_node = start_node
        segment.end_node = end_node
        segment.relationship = relationship
        return segment


class Node(BaseObject):
    def __init__(self, vertex):
        super(Node, self).__init__()
        self._value = vertex
        self._tag_indexes = dict()
        for index, tag in enumerate(self._value.tags, start=0):
            self._tag_indexes[tag.name.decode(self.get_decode_type())] = index

    def get_id(self):
        """get the vid of Node

        :return: ValueWrapper type vid
        """
        return ValueWrapper(value=self._value.vid,
                            decode_type=self.get_decode_type(),
                            timezone_offset=self.get_timezone_offset())

    def tags(self):
        """get tag names

        :return: the list of tag name
        """
        return list(self._tag_indexes.keys())

    def has_tag(self, tag):
        """whether the specified tag is included

        :param tag: the tag name
        :return: true or false
        """
        return True if tag in self._tag_indexes.keys() else False

    def properties(self, tag):
        """get all properties of the specified tag

        :param tag: the tag name
        :return: the properties
        """
        if tag not in self._tag_indexes.keys():
            raise InvalidKeyException(tag)

        props = self._value.tags[self._tag_indexes[tag]].props
        result_props = {}
        if props is None:
            return result_props
        for key in props.keys():
            result_props[key.decode(self.get_decode_type())] = ValueWrapper(props[key],
                                                                            decode_type=self.get_decode_type(),
                                                                            timezone_offset=self._timezone_offset)
        return result_props

    def prop_names(self, tag):
        """get the property names of the specified tag

        :param tag: the tag name
        :return: property name list
        """
        if tag not in self._tag_indexes.keys():
            raise InvalidKeyException(tag)
        index = self._tag_indexes[tag]
        props = self._value.tags[index].props
        if props is None:
            return []
        return [(key.decode(self.get_decode_type())) for key in self._value.tags[index].props.keys()]

    def prop_values(self, tag):
        """get all property values of the specified tag

        :param tag: the tag name
        :return: property name list
        """
        if tag not in self._tag_indexes.keys():
            raise InvalidKeyException(tag)
        index = self._tag_indexes[tag]
        props = self._value.tags[index].props
        if props is None:
            return []
        return [(ValueWrapper(value,
                              decode_type=self.get_decode_type(),
                              timezone_offset=self._timezone_offset))
                for value in self._value.tags[index].props.values()]

    def __repr__(self):
        tag_str_list = list()
        for tag in self._tag_indexes.keys():
            prop_strs = ['%s: %s' % (key, str(val))
                         for key, val in self.properties(tag).items()]
            tag_str_list.append(':%s{%s}' % (tag, ', '.join(prop_strs)))
        return '({} {})'.format(str(self.get_id()), ' '.join(tag_str_list))

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False

        return self.get_id() == other.get_id()

    def __ne__(self, other):
        return not (self == other)


class Relationship(BaseObject):
    def __init__(self, edge: Edge):
        super(Relationship, self).__init__()
        self._value = edge

    def start_vertex_id(self):
        """get start vertex vid, if your space vid_type is int, you can use start_vertex_id().as_int(),
        if your space vid_type is fixed_string, you can use start_vertex_id().as_string()

        :return: ValueWrapper type vid
        """
        if self._value.type > 0:
            return ValueWrapper(self._value.src, self.get_decode_type())
        else:
            return ValueWrapper(self._value.dst, self.get_decode_type())

    def end_vertex_id(self):
        """get end vertex vid, if your space vid_type is int, you can use end_vertex_id().as_int(),
        if your space vid_type is fixed_string, you can use end_vertex_id().as_string()

        :return: ValueWrapper type vid
        """
        if self._value.type > 0:
            return ValueWrapper(self._value.dst, self.get_decode_type())
        else:
            return ValueWrapper(self._value.src, self.get_decode_type())

    def edge_name(self):
        """get the edge name

        :return: edge name
        """
        return self._value.name.decode(self.get_decode_type())

    def ranking(self):
        """get the edge ranking

        :return: ranking
        """
        return self._value.ranking

    def properties(self):
        """get all properties

        :return: the properties
        """
        props = {}
        if self._value.props is None:
            return props
        for key in self._value.props.keys():
            props[key.decode(self.get_decode_type())] = ValueWrapper(self._value.props[key],
                                                                     decode_type=self.get_decode_type(),
                                                                     timezone_offset=self.get_timezone_offset())
        return props

    def keys(self):
        """get all property names

        :return: the property names
        """
        if self._value.props is None:
            return []
        return [(key.decode(self._decode_type)) for key in self._value.props.keys()]

    def values(self):
        """get all property values

        :return: the property values
        """
        if self._value.props is None:
            return []
        return [(ValueWrapper(value,
                              decode_type=self.get_decode_type(),
                              timezone_offset=self.get_timezone_offset()))
                for value in self._value.props.values()]

    def __repr__(self):
        prop_strs = ['%s: %s' % (key, str(val))
                     for key, val in self.properties().items()]
        return "(%s)-[:%s@%d{%s}]->(%s)" % (str(self.start_vertex_id()),
                                            self.edge_name(),
                                            self.ranking(),
                                            ', '.join(prop_strs),
                                            str(self.end_vertex_id()))

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return self.start_vertex_id() == other.start_vertex_id() \
            and self.end_vertex_id() == other.end_vertex_id() \
            and self.edge_name() == other.edge_name() \
            and self.ranking() == self.ranking()

    def __ne__(self, other):
        return not (self == other)


class Segment:
    start_node = None
    end_node = None
    relationship = None

    def __repr__(self):
        return "{}-[:{}@{}{}]->{}".format(self.start_node,
                                          self.relationship.edge_name(),
                                          self.relationship.ranking(),
                                          self.relationship.properties(),
                                          self.end_node)

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return self.start_node == other.start_node \
            and self.end_node == other.end_node \
            and self.relationship == other.relationship


class PathWrapper(BaseObject):
    """
    PathWrapper is wrapper handling for the Path from the service
    """

    def __init__(self, path):
        super(PathWrapper, self).__init__()
        self._nodes = list()
        self._segments = list()
        self._relationships = list()

        self._path = path
        self._nodes.append(Node(path.src)
                           .set_decode_type(self.get_decode_type())
                           .set_timezone_offset(self.get_timezone_offset()))

        vids = []
        vids.append(path.src.vid)
        for step in self._path.steps:
            type = step.type
            if step.type > 0:
                start_node = self._nodes[-1]
                end_node = Node(step.dst)\
                    .set_decode_type(self.get_decode_type())\
                    .set_timezone_offset(self.get_timezone_offset())
                src_id = vids[-1]
                dst_id = step.dst.vid
            else:
                type = -type
                end_node = self._nodes[-1]
                start_node = Node(step.dst)\
                    .set_decode_type(self.get_decode_type())\
                    .set_timezone_offset(self.get_timezone_offset())
                dst_id = vids[-1]
                src_id = step.dst.vid
            vids.append(step.dst.vid)
            relationship = Relationship(GenValue.gen_edge(src_id,
                                                          dst_id,
                                                          type,
                                                          step.name,
                                                          step.ranking,
                                                          step.props))\
                .set_decode_type(self.get_decode_type())\
                .set_timezone_offset(self.get_timezone_offset())

            self._relationships.append(relationship)
            segment = GenValue.gen_segment(start_node, end_node, relationship)
            if segment.start_node == self._nodes[-1]:
                self._nodes.append(segment.end_node)
            elif segment.end_node == self._nodes[-1]:
                self._nodes.append(segment.start_node)
            else:
                raise Exception("Relationship [{}] does not connect to the last node".
                                format(relationship))

            self._segments.append(segment)

    def __iter__(self):
        return iter(self._segments)

    def start_node(self):
        """get start node of the Path

        :return: start node
        """
        if len(self._nodes) == 0:
            return None
        return self._nodes[0]

    def length(self):
        """get the length of the path

        :return: path length
        """
        return len(self._segments)

    def contain_node(self, node):
        """whether the node is in the path

        :param node: the specified node
        :return: true or false
        """
        return True if node in self._nodes else False

    def contain_relationship(self, relationship):
        """whether the relationship is in the path

        :param relationship: the specified relationship
        :return: true or false
        """
        return True if relationship in self._relationships else False

    def nodes(self):
        """get all nodes of the path

        :return: nodes
        """
        return self._nodes

    def relationships(self):
        """get all relationships of the path

        :return: relationships
        """
        return self._relationships

    def segments(self):
        """get all segments of the path

        :return: segments
        """
        return self._segments

    def __repr__(self):
        edge_strs = []
        for step in self._path.steps:
            relationship = Relationship(GenValue.gen_edge(step.dst.vid,
                                                          step.dst.vid,
                                                          type,
                                                          step.name,
                                                          step.ranking,
                                                          step.props))\
                .set_decode_type(self.get_decode_type())\
                .set_timezone_offset(self.get_timezone_offset())
            edge_str = ''
            prop_strs = ['%s: %s' % (key, str(val))
                         for key, val in relationship.properties().items()]
            if step.type > 0:
                edge_str = '-[:%s@%d{%s}]->%s' % (relationship.edge_name(),
                                                  relationship.ranking(),
                                                  ', '.join(prop_strs),
                                                  Node(step.dst)
                                                  .set_decode_type(self.get_decode_type())
                                                  .set_timezone_offset(self.get_timezone_offset()))
            else:
                edge_str = "<-[:%s@%d{%s}]-%s" % (relationship.edge_name(),
                                                  relationship.ranking(),
                                                  ', '.join(prop_strs),
                                                  Node(step.dst)
                                                  .set_decode_type(self.get_decode_type())
                                                  .set_timezone_offset(self.get_timezone_offset())
                                                  )

            edge_strs.append(edge_str)
        return '{}{}'.format(Node(self._path.src)
                             .set_decode_type(self.get_decode_type())
                             .set_timezone_offset(self.get_timezone_offset()), ''.join(edge_strs))

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False

        return self._segments == other.segments()

    def __ne__(self, other):
        return not (self == other)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from nebula2.fbthrift.transport.TSocket import TSocket
from nebula2.fbthrift.transport.TTransport import TTransportException
import socket


class TSocketOverHttpTunnel(TSocket):
    def __init__(self, host, port, proxy_host, proxy_port):
        TSocket.__init__(self, proxy_host, proxy_port)
        try:
            # Use IP address since sometimes proxy_host cannot resolve
            # external hostnames using unbound
            info = socket.getaddrinfo(
                host,
                None,
                socket.AF_INET | socket.AF_INET6,
                socket.SOCK_STREAM,
                socket.IPPROTO_TCP)
            self.remote_host = info[0][4][0]
        except socket.error as e:
            raise TTransportException(TTransportException.NOT_OPEN, str(e))
        self.remote_port = port

    def open(self):
        TSocket.open(self)
        self.write("CONNECT %s:%d HTTP/1.1\r\nHost: %s:%d\r\n\r\n" % (
            self.remote_host, self.remote_port,
            self.remote_host, self.remote_port))
        res = self.read(4096)
        try:
            status = res.split()[1]
            if status != '200':
                self.close()
                raise TTransportException(TTransportException.NOT_OPEN,
                        "Error response from proxy server: %s" % res)
        except IndexError:
            self.close()
            raise TTransportException(TTransportException.NOT_OPEN,
                    "Error response from proxy server: %s" % res)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import socket
import ssl

from nebula2.fbthrift.transport.TSocketOverHttpTunnel import TSocketOverHttpTunnel
from nebula2.fbthrift.transport.TTransport import TTransportException

class TSSLSocketOverHttpTunnel(TSocketOverHttpTunnel):
    def __init__(self, host, port, proxy_host, proxy_port,
                 ssl_version=ssl.PROTOCOL_TLSv1,
                 cert_reqs=ssl.CERT_NONE,
                 ca_certs=None,
                 keyfile=None,
                 certfile=None):
        TSocketOverHttpTunnel.__init__(self, host, port, proxy_host, proxy_port)
        self.ssl_version = ssl_version
        self.cert_reqs = cert_reqs
        self.keyfile, self.certfile, self.ca_certs = \
                keyfile, certfile, ca_certs

    def open(self):
        TSocketOverHttpTunnel.open(self)
        try:
            sslh = ssl.SSLSocket(self.handle,
                                 ssl_version=self.ssl_version,
                                 cert_reqs=self.cert_reqs,
                                 keyfile=self.keyfile,
                                 certfile=self.certfile,
                                 ca_certs=self.ca_certs)
            self.handle = sslh
        except ssl.SSLError as e:
            self.close()
            raise TTransportException(TTransportException.NOT_OPEN,
                    "SSL error during handshake: " + str(e))
        except socket.error as e:
            self.close()
            raise TTransportException(TTransportException.NOT_OPEN,
                    "socket error during SSL handshake: " + str(e))
